/*
 Mini Temp Logger
 Version: v1.1
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 22/02/2015
 Last Modified: 4/04/2017
 
 */

#define LOW 0
#define HIGH 1
#define false 0
#define true 1

// this is libusb, see http://libusb.sourceforge.net/ 
#include <usb.h>

// USB functions
#define USB_TRANSFER 1
#define USB_DATA_FROM_PC 2
#define USB_DELAY_TIME 3
#define USB_RTC_START_LOG_TIME 4
#define USB_RTC_CURRENT_TIME 5
#define USB_RTC_TRIM 6
#define USB_TEMP 7
#define USB_AUTOTRIM 8
#define USB_AUTOTRIM_VAL_AND_CALC 9
#define USB_AUTOTRIM_INTERVAL 11
#define USB_FUNCTION_SELECTED 12
#define USB_VOLTAGE 13

// USB sub functions when sending data from the PC
#define PROGRAM_DELAY_TIME 1
#define PROGRAM_RTC_TIME 2
#define PROGRAM_RTC_TRIM 3
#define PROGRAM_AUTOTRIM 4
#define PROGRAM_AUTOTRIM_INTERVAL 5
#define PROGRAM_LOG_FUNCTION 6

#define LOG_TEMPERATURE 1
#define LOG_VOLTAGE 2

#define endLoggingNumber 57081 // If the EEPROM bytes combine to form this number, this is the end of the data

// Global variables
usb_dev_handle *handle = NULL;

char buffer[256];
char verifyBuffer[256];

int days = 0;
int hours = 0;
int minutes = 0;
int seconds = 0;

double voltageDividerMultipler = 0;
float voltageCalibration = 0;


// Read the config.ini file for the voltage divider / calibration
void read_config(void) {
	FILE* configfile = fopen ( "config.ini" , "rb" );
	char* buffer;
	if (configfile != NULL) {
		// Obtain file size
		fseek (configfile , 0 , SEEK_END);
		long fileSize = ftell (configfile);
		rewind (configfile);
		
		// Allocate memory 
		buffer = (char*) malloc (sizeof(char) * fileSize);
		
		// Copy the file into the buffer
		fread (buffer, 1, fileSize, configfile);
		
		// Break up 2 lines
		char *pch;
		pch = strtok (buffer,"\r\n");
		if (pch != NULL) {
			voltageDividerMultipler = atof(pch);
			
			pch = strtok (NULL, "\r\n");
			if (pch != NULL) {
				voltageCalibration = atof(pch);
			}
		}
		
		fclose (configfile);
		//free (buffer);
	}
}

// Write the config.ini file for the voltage divider / calibration
void write_config(void) {
	FILE* configfile = fopen("config.ini", "wb");

	if (configfile != NULL) {
		char voltageDividerMultiplerBuffer[10];
		snprintf(voltageDividerMultiplerBuffer, 9, "%.3f", voltageDividerMultipler);
		//printf("%s\n", voltageDividerMultiplerBuffer);
		
		char voltageCalibrationBuffer[20];
		snprintf(voltageCalibrationBuffer, 19, "%.11f", voltageCalibration);
		//printf("%s\n", voltageCalibrationBuffer);
		
		fwrite(voltageDividerMultiplerBuffer, 1, strlen(voltageDividerMultiplerBuffer), configfile);
		fwrite("\r\n", 1, 2, configfile);
		fwrite(voltageCalibrationBuffer, 1, strlen(voltageCalibrationBuffer), configfile);
		fwrite("\r\n", 1, 2, configfile);
		
		fclose(configfile);
	}
}

uint8_t read_number_w_min_max(uint8_t min, uint8_t max) {
	int number = 0;
	scanf("%d", &number);
	while (number < min || number > max) {
		printf("\nInvalid, please try again: ");
		scanf("%d", &number);
	}
	
	return number;
}

// Request input of days, hours, minutes, seconds
void input_dhms(void) {
	printf("Please enter in the days: ");
	days = read_number_w_min_max(0, 255);
	
	printf("Please enter in the hours: ");
	hours = read_number_w_min_max(0, 24);
	
	printf("Please enter in the minutes: ");
	minutes = read_number_w_min_max(0, 60);
	
	printf("Please enter in the seconds: ");
	seconds = read_number_w_min_max(0, 60);
}

// Print the time from the read buffer
void print_time_from_buffer(void) {
	uint8_t readRtcAmPm = buffer[0];
	uint8_t readRtcSeconds = buffer[1];
	uint8_t readRtcMinutes = buffer[2];
	uint8_t readRtcHour = buffer[3];
	uint8_t readRtcDate = buffer[4];
	uint8_t readRtcMonth = buffer[5];
	uint8_t readRtcYear = buffer[6];
	
	printf("Year = 20%i\n", readRtcYear);
	printf("Month = %i\n", readRtcMonth);
	printf("Date = %i\n", readRtcDate);
	printf("Hour = %i\n", readRtcHour);
	printf("Minutes = %i\n", readRtcMinutes);
	printf("Seconds = %i\n", readRtcSeconds);
	printf("Am/Pm = ");
	if (readRtcAmPm == 0) {
		printf("AM\n\n");
	}
	else {
		printf("PM\n\n");
	}
}


// Print the interval time from the read buffer
void print_interval_time_from_buffer(void) {
	uint8_t readDelayTimeSeconds = buffer[0];
	uint8_t readDelayTimeMinutes = buffer[1];
	uint8_t readDelayTimeHours = buffer[2];
	uint8_t readDelayTimeDays = buffer[3];
	
	printf("Days = %i\n", readDelayTimeDays);
	printf("Hours = %i\n", readDelayTimeHours);
	printf("Minutes = %i\n", readDelayTimeMinutes);
	printf("Seconds = %i\n\n", readDelayTimeSeconds);
}

// Clear buffer
void clear_buffer(void) {
	for (uint8_t x = 0; x < 255; x++) {
		buffer[x] = 0;
	}
}

// Send buffer with the data type
void usb_send_data(uint8_t type, uint8_t bufferSize) {
	buffer[0] = type;
	
	usb_control_msg(handle, 
		USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT, 
		USB_DATA_FROM_PC, 0, 0, buffer, (bufferSize+2), 5000); // Include type +1 and null character +1 at the end in the buffer size
	
	// Wait a little bit
	Sleep(500);
}

// Read from the USB device with the data type, clear the global buffer first
void usb_read_data(uint8_t type) {
	clear_buffer();
	
	int nBytes = usb_control_msg(handle, 
					USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
					type, 0, 0, (char *)buffer, sizeof(buffer), 5000);
	
	if (nBytes == 0) {
		printf("Error: Failed to receive data\n");
	}
}

// Verify the data we wrote on the global buffer against the read we will perform now
void verify_data_written(uint8_t type, uint8_t bufferSize) {
	int nBytes = usb_control_msg(handle, 
					USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
					type, 0, 0, (char *)verifyBuffer, sizeof(verifyBuffer), 5000);
	
	if (nBytes == 0) {
		printf("Error: Failed to receive data\n");
	}
	
	// Verify the bytes, start from 1 on buffer and 0 on verify buffer
	uint8_t verifiedOk = true;
	for (uint8_t x = 0; x <= bufferSize; x++) {
		if (buffer[x+1] != verifyBuffer[x]) {
			verifiedOk = false;
		}
	}
	
	if (verifiedOk == true) {
		printf("Completed\n");
	}
	else {
		printf("\nError: Programmed values didn't match\n");
		printf("Local\t\tUSB returned\n");
		
		for (uint8_t x = 0; x <= bufferSize; x++) {
			printf("%i\t\t%i\n", (uint8_t) buffer[x+1], (uint8_t) verifyBuffer[x]);
		}
	}
}

void check_arguments(int argc, char **argv) {
	// Arguments from GUI
	if (argc >= 2) {
		if (strstr(argv[1], "/check")) {
			printf("1");
			
			usb_close(handle);
			exit(0);
		}
		else if (strstr(argv[1], "/pdt")) {	// Program delay time
			/*printf("%s ", argv[2]);
			printf("%s ", argv[3]);
			printf("%s ", argv[4]);
			printf("%s ", argv[5]);*/
			
			seconds = atoi(argv[2]);
			minutes = atoi(argv[3]);
			hours = atoi(argv[4]);
			days = atoi(argv[5]);
			
			// Send to USB device
			clear_buffer();
			buffer[1] = seconds;
			buffer[2] = minutes;
			buffer[3] = hours;
			buffer[4] = days;
			usb_send_data(PROGRAM_DELAY_TIME, 4);
			verify_data_written(USB_DELAY_TIME, 4);
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/rdt")) { // Read delay time
			usb_read_data(USB_DELAY_TIME);
			
			uint8_t readDelayTimeSeconds = buffer[0];
			uint8_t readDelayTimeMinutes = buffer[1];
			uint8_t readDelayTimeHours = buffer[2];
			uint8_t readDelayTimeDays = buffer[3];
			
			printf("%i %i %i %i", readDelayTimeSeconds, readDelayTimeMinutes, readDelayTimeHours, readDelayTimeDays);
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/plf")) {
			clear_buffer();
			buffer[1] = atoi(argv[2]);
			usb_send_data(PROGRAM_LOG_FUNCTION, 1);
			verify_data_written(USB_FUNCTION_SELECTED, 1);
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/rlf")) {
			usb_read_data(USB_FUNCTION_SELECTED);
			printf("%i", buffer[0]);
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/readrtcseparate")) {
			usb_read_data(USB_RTC_CURRENT_TIME);
			
			uint8_t readRtcAmPm = buffer[0];
			uint8_t readRtcSeconds = buffer[1];
			uint8_t readRtcMinutes = buffer[2];
			uint8_t readRtcHour = buffer[3];
			uint8_t readRtcDate = buffer[4];
			uint8_t readRtcMonth = buffer[5];
			uint8_t readRtcYear = buffer[6];
			
			printf("20%i ", readRtcYear);
			
			if (readRtcMonth < 10) {
				printf("0%i ", readRtcMonth);
			}
			else {
				printf("%i ", readRtcMonth);
			}
			
			if (readRtcDate < 10) {
				printf("0%i ", readRtcDate);
			}
			else {
				printf("%i ", readRtcDate);
			}
			
			if (readRtcHour < 10) {
				printf("0%i ", readRtcHour);
			}
			else {
				printf("%i ", readRtcHour);
			}
			
			if (readRtcMinutes < 10) {
				printf("0%i ", readRtcMinutes);
			}
			else {
				printf("%i ", readRtcMinutes);
			}
			
			if (readRtcSeconds < 10) {
				printf("0%i ", readRtcSeconds);
			}
			else {
				printf("%i ", readRtcSeconds);
			}
			
			if (readRtcAmPm == 0) {
				printf("AM");
			}
			else {
				printf("PM");
			}
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/readrtctime")) {
			usb_read_data(USB_RTC_CURRENT_TIME);
			
			uint8_t readRtcAmPm = buffer[0];
			uint8_t readRtcSeconds = buffer[1];
			uint8_t readRtcMinutes = buffer[2];
			uint8_t readRtcHour = buffer[3];
			uint8_t readRtcDate = buffer[4];
			uint8_t readRtcMonth = buffer[5];
			uint8_t readRtcYear = buffer[6];
			
			printf("20%i-", readRtcYear);
			
			if (readRtcMonth < 10) {
				printf("0%i-", readRtcMonth);
			}
			else {
				printf("%i-", readRtcMonth);
			}
			
			if (readRtcDate < 10) {
				printf("0%i", readRtcDate);
			}
			else {
				printf("%i", readRtcDate);
			}
			
			if (readRtcHour < 10) {
				printf(" 0%i:", readRtcHour);
			}
			else {
				printf(" %i:", readRtcHour);
			}
			
			if (readRtcMinutes < 10) {
				printf("0%i:", readRtcMinutes);
			}
			else {
				printf("%i:", readRtcMinutes);
			}
			
			if (readRtcSeconds < 10) {
				printf("0%i ", readRtcSeconds);
			}
			else {
				printf("%i ", readRtcSeconds);
			}
			
			if (readRtcAmPm == 0) {
				printf("AM");
			}
			else {
				printf("PM");
			}
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/readlogstart")) {
			usb_read_data(USB_RTC_START_LOG_TIME);
			
			uint8_t readRtcAmPm = buffer[0];
			uint8_t readRtcSeconds = buffer[1];
			uint8_t readRtcMinutes = buffer[2];
			uint8_t readRtcHour = buffer[3];
			uint8_t readRtcDate = buffer[4];
			uint8_t readRtcMonth = buffer[5];
			uint8_t readRtcYear = buffer[6];
			
			printf("20%i-", readRtcYear);
			
			if (readRtcMonth < 10) {
				printf("0%i-", readRtcMonth);
			}
			else {
				printf("%i-", readRtcMonth);
			}
			
			if (readRtcDate < 10) {
				printf("0%i", readRtcDate);
			}
			else {
				printf("%i", readRtcDate);
			}
			
			if (readRtcHour < 10) {
				printf(" 0%i:", readRtcHour);
			}
			else {
				printf(" %i:", readRtcHour);
			}
			
			if (readRtcMinutes < 10) {
				printf("0%i:", readRtcMinutes);
			}
			else {
				printf("%i:", readRtcMinutes);
			}
			
			if (readRtcSeconds < 10) {
				printf("0%i ", readRtcSeconds);
			}
			else {
				printf("%i ", readRtcSeconds);
			}
			
			if (readRtcAmPm == 0) {
				printf("AM");
			}
			else {
				printf("PM");
			}
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/readrtctrim")) {
			usb_read_data(USB_RTC_TRIM);
			
			uint8_t readRtcTrim = buffer[0];
			printf("%i", readRtcTrim & 0x7F);
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/readrtcpm")) {
			usb_read_data(USB_RTC_TRIM);
			
			uint8_t readRtcTrim = buffer[0];
			if (readRtcTrim & 1<<7) {
				printf("+");
			}
			else {
				printf("-");
			}
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/readrtcautotrim")) {
			usb_read_data(USB_AUTOTRIM);
			
			printf("%i", buffer[0]);
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/readautotrimppm")) {
			usb_read_data(USB_AUTOTRIM_VAL_AND_CALC);		
			
			uint8_t readRtcAutoTrimVal = buffer[0];
			printf("%i", readRtcAutoTrimVal);
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/readautotrimapplied")) {
			usb_read_data(USB_AUTOTRIM_VAL_AND_CALC);		
			
			uint8_t readRtcAutoTrimCalc = buffer[1];
			if (readRtcAutoTrimCalc & 1<<7) {
				printf("+%i PPM (%i clock cycles)", readRtcAutoTrimCalc & 0x7F, (readRtcAutoTrimCalc & 0x7F) * 2);
			}
			else {
				printf("-%i PPM (%i clock cycles)", readRtcAutoTrimCalc & 0x7F, (readRtcAutoTrimCalc & 0x7F) * 2);
			}
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/readautotriminterval")) {
			usb_read_data(USB_AUTOTRIM_INTERVAL);
			
			uint8_t readDelayTimeSeconds = buffer[0];
			uint8_t readDelayTimeMinutes = buffer[1];
			uint8_t readDelayTimeHours = buffer[2];
			uint8_t readDelayTimeDays = buffer[3];
			
			printf("%i ", readDelayTimeDays);
			printf("%i ", readDelayTimeHours);
			printf("%i ", readDelayTimeMinutes);
			printf("%i", readDelayTimeSeconds);
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/readtemp")) {
			usb_read_data(USB_TEMP);
			
			int combinedValue = ((uint8_t) buffer[0] << 8) | (uint8_t) buffer[1];
			if (combinedValue >= 0x1C90 && combinedValue <= 0x1FFF) { // Convert negative temps
				combinedValue &= 0x1FF;
				combinedValue--;
				combinedValue ^= 0x1FF;
				printf("-");
			}
			printf("%4.2f C", combinedValue * 0.0625);
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/readvolt")) {
			read_config();
			usb_read_data(USB_VOLTAGE);
			
			int combinedValue = ((uint8_t) buffer[0] << 8) | (uint8_t) buffer[1];
			printf("%2.3f V", (double) combinedValue * (0.00107421875 * voltageDividerMultipler) - (combinedValue * voltageCalibration)); // voltageCalibration is the calculated difference between the actual ADC result and what the battery really measures
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/rtcprogramtime")) {
			time_t rawtime;
			struct tm* timeinfo;
			char timebuffer[80];
			
			time (&rawtime);
			timeinfo = localtime (&rawtime);
			strftime (timebuffer,80,"%Y-%m-%d-%H-%M-%S",timeinfo);
			
			int tokenCount = 7;
			char *token;
			token = strtok(timebuffer, "-");
			while(token != NULL) {
				if (tokenCount == 7) { // Year
					buffer[tokenCount] = atoi(token) - 2000;
				}
				else {
					buffer[tokenCount] = atoi(token);
				}
				
				token = strtok(NULL, "-");
				tokenCount--;
			}
			
			// Convert to 12 hour format
			buffer[1] = 0; // Am
			if (buffer[4] >= 13) {
				buffer[4] -= 12;
				buffer[1] = 1; // Pm
			}
			else if (buffer[4] == 12) {
				buffer[1] = 1; // Pm
			}
			else if (buffer[4] == 0) {
				buffer[4] = 12;
			}
			
			usb_send_data(PROGRAM_RTC_TIME, 7);
			verify_data_written(USB_RTC_CURRENT_TIME, 7);
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/rtcprogrammanual")) {
			uint8_t AmPm = atoi(argv[2]);
			uint8_t seconds = atoi(argv[3]);
			uint8_t minutes = atoi(argv[4]);
			uint8_t hours = atoi(argv[5]);
			uint8_t date = atoi(argv[6]);
			uint8_t month = atoi(argv[7]);
			uint8_t year = atoi(argv[8]);
			
			// Send to USB device
			buffer[1] = AmPm;
			buffer[2] = seconds;
			buffer[3] = minutes;
			buffer[4] = hours;
			buffer[5] = date;
			buffer[6] = month;
			buffer[7] = year;
			usb_send_data(PROGRAM_RTC_TIME, 7);
			verify_data_written(USB_RTC_CURRENT_TIME, 7);
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/prtctrim")) {
			int rtcTrim = atoi(argv[3]);
			int rtcTrimSign = atoi(argv[2]);
			
			rtcTrim |= (rtcTrimSign << 7);
			
			buffer[1] = rtcTrim;
			usb_send_data(PROGRAM_RTC_TRIM, 1);
			verify_data_written(USB_RTC_TRIM, 1);
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/pautortctrim")) {
			buffer[1] = atoi(argv[2]);
			usb_send_data(PROGRAM_AUTOTRIM, 1);
			verify_data_written(USB_AUTOTRIM, 1);
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/pati")) {
			seconds = atoi(argv[2]);
			minutes = atoi(argv[3]);
			hours = atoi(argv[4]);
			days = atoi(argv[5]);
			
			buffer[1] = seconds;
			buffer[2] = minutes;
			buffer[3] = hours;
			buffer[4] = days;
			usb_send_data(PROGRAM_AUTOTRIM_INTERVAL, 4);
			verify_data_written(USB_AUTOTRIM_INTERVAL, 4);
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/refresh")) {
			read_config();
			usb_read_data(USB_TEMP);
			
			int combinedValue = ((uint8_t) buffer[0] << 8) | (uint8_t) buffer[1];
			if (combinedValue >= 0x1C90 && combinedValue <= 0x1FFF) { // Convert negative temps
				combinedValue &= 0x1FF;
				combinedValue--;
				combinedValue ^= 0x1FF;
				printf("-");
			}
			printf("%4.2f C*", combinedValue * 0.0625);
			
			
			usb_read_data(USB_VOLTAGE);
			
			combinedValue = ((uint8_t) buffer[0] << 8) | (uint8_t) buffer[1];
			printf("%2.3f V*", (double) combinedValue * (0.00107421875 * voltageDividerMultipler) - (combinedValue * voltageCalibration)); // voltageCalibration is the calculated difference between the actual ADC result and what the battery really measures
			
			
			usb_read_data(USB_RTC_CURRENT_TIME);
			
			uint8_t readRtcAmPm = buffer[0];
			uint8_t readRtcSeconds = buffer[1];
			uint8_t readRtcMinutes = buffer[2];
			uint8_t readRtcHour = buffer[3];
			uint8_t readRtcDate = buffer[4];
			uint8_t readRtcMonth = buffer[5];
			uint8_t readRtcYear = buffer[6];
			
			printf("20%i-", readRtcYear);
			
			if (readRtcMonth < 10) {
				printf("0%i-", readRtcMonth);
			}
			else {
				printf("%i-", readRtcMonth);
			}
			
			if (readRtcDate < 10) {
				printf("0%i", readRtcDate);
			}
			else {
				printf("%i", readRtcDate);
			}
			
			if (readRtcHour < 10) {
				printf(" 0%i:", readRtcHour);
			}
			else {
				printf(" %i:", readRtcHour);
			}
			
			if (readRtcMinutes < 10) {
				printf("0%i:", readRtcMinutes);
			}
			else {
				printf("%i:", readRtcMinutes);
			}
			
			if (readRtcSeconds < 10) {
				printf("0%i ", readRtcSeconds);
			}
			else {
				printf("%i ", readRtcSeconds);
			}
			
			if (readRtcAmPm == 0) {
				printf("AM*");
			}
			else {
				printf("PM*");
			}
			
			
			usb_read_data(USB_AUTOTRIM_VAL_AND_CALC);	
			
			uint8_t readRtcAutoTrimVal = buffer[0];
			printf("-%i PPM*", readRtcAutoTrimVal);	
			
			uint8_t readRtcAutoTrimCalc = buffer[1];
			if (readRtcAutoTrimCalc & 1<<7) {
				printf("+%i PPM (%i clock cycles)", readRtcAutoTrimCalc & 0x7F, (readRtcAutoTrimCalc & 0x7F) * 2);
			}
			else {
				printf("-%i PPM (%i clock cycles)", readRtcAutoTrimCalc & 0x7F, (readRtcAutoTrimCalc & 0x7F) * 2);
			}
			
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/readconfig")) {
			read_config();
			printf("%.2f*", voltageDividerMultipler);
			printf("%.11f", voltageCalibration);
			usb_close(handle);
			exit(1);
		}
		else if (strstr(argv[1], "/writeconfig")) {
			voltageDividerMultipler = atof(argv[2]);
			voltageCalibration = atof(argv[3]);
			
			write_config();
			usb_close(handle);
			exit(1);
		}
	}
}


// used to get descriptor strings for device identification 
static int usbGetDescriptorString(usb_dev_handle *dev, int index, int langid, 
                                  char *buf, int buflen) {
    char buffer[256];
    int rval, i;

	// make standard request GET_DESCRIPTOR, type string and given index 
    // (e.g. dev->iProduct)
	rval = usb_control_msg(dev, 
        USB_TYPE_STANDARD | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 
        USB_REQ_GET_DESCRIPTOR, (USB_DT_STRING << 8) + index, langid, 
        buffer, sizeof(buffer), 1000);
        
    if(rval < 0) // error
		return rval;
	
    // rval should be bytes read, but buffer[0] contains the actual response size
	if((unsigned char)buffer[0] < rval)
		rval = (unsigned char)buffer[0]; // string is shorter than bytes read
	
	if(buffer[1] != USB_DT_STRING) // second byte is the data type
		return 0; // invalid return type
		
	// we're dealing with UTF-16LE here so actual chars is half of rval,
	// and index 0 doesn't count
	rval /= 2;
	
	// lossy conversion to ISO Latin1 
	for(i = 1; i < rval && i < buflen; i++) {
		if(buffer[2 * i + 1] == 0)
			buf[i-1] = buffer[2 * i];
		else
			buf[i-1] = '?'; // outside of ISO Latin1 range
	}
	buf[i-1] = 0;
	
	return i-1;
}

static usb_dev_handle * usbOpenDevice(int vendor, char *vendorName, 
                                      int product, char *productName) {
	struct usb_bus *bus;
	struct usb_device *dev;
	char devVendor[256], devProduct[256];
    
	usb_dev_handle * handle = NULL;
	
	usb_init();
	usb_find_busses();
	usb_find_devices();
	
	for(bus=usb_get_busses(); bus; bus=bus->next) {
		for(dev=bus->devices; dev; dev=dev->next) {			
			if(dev->descriptor.idVendor != vendor ||
               dev->descriptor.idProduct != product)
                continue;
                
            // we need to open the device in order to query strings 
            if(!(handle = usb_open(dev))) {
                fprintf(stderr, "Warning: cannot open USB device: %s\n",
                    usb_strerror());
                continue;
            }
            
            // get vendor name 
            if(usbGetDescriptorString(handle, dev->descriptor.iManufacturer, 0x0409, devVendor, sizeof(devVendor)) < 0) {
                fprintf(stderr, 
                    "Warning: cannot query manufacturer for device: %s\n", 
                    usb_strerror());
                usb_close(handle);
                continue;
            }
            
            // get product name 
            if(usbGetDescriptorString(handle, dev->descriptor.iProduct, 
               0x0409, devProduct, sizeof(devVendor)) < 0) {
                fprintf(stderr, 
                    "Warning: cannot query product for device: %s\n", 
                    usb_strerror());
                usb_close(handle);
                continue;
            }
            
            if(strcmp(devVendor, vendorName) == 0 && 
               strcmp(devProduct, productName) == 0)
                return handle;
            else
                usb_close(handle);
		}
	}
	
	return NULL;
}
